// swift-tools-version: 5.7

import PackageDescription

let package = Package(
    name: "HelloAlgo",
    products: [
        // chapter_computational_complexity
        .executable(name: "iteration", targets: ["iteration"]),
        .executable(name: "recursion", targets: ["recursion"]),
        .executable(name: "time_complexity", targets: ["time_complexity"]),
        .executable(name: "worst_best_time_complexity", targets: ["worst_best_time_complexity"]),
        .executable(name: "space_complexity", targets: ["space_complexity"]),
        // chapter_array_and_linkedlist
        .executable(name: "array", targets: ["array"]),
        .executable(name: "linked_list", targets: ["linked_list"]),
        .executable(name: "list", targets: ["list"]),
        .executable(name: "my_list", targets: ["my_list"]),
        // chapter_stack_and_queue
        .executable(name: "stack", targets: ["stack"]),
        .executable(name: "linkedlist_stack", targets: ["linkedlist_stack"]),
        .executable(name: "array_stack", targets: ["array_stack"]),
        .executable(name: "queue", targets: ["queue"]),
        .executable(name: "linkedlist_queue", targets: ["linkedlist_queue"]),
        .executable(name: "array_queue", targets: ["array_queue"]),
        .executable(name: "deque", targets: ["deque"]),
        .executable(name: "linkedlist_deque", targets: ["linkedlist_deque"]),
        .executable(name: "array_deque", targets: ["array_deque"]),
        // chapter_hashing
        .executable(name: "hash_map", targets: ["hash_map"]),
        .executable(name: "array_hash_map", targets: ["array_hash_map"]),
        .executable(name: "hash_map_chaining", targets: ["hash_map_chaining"]),
        .executable(name: "hash_map_open_addressing", targets: ["hash_map_open_addressing"]),
        .executable(name: "simple_hash", targets: ["simple_hash"]),
        .executable(name: "built_in_hash", targets: ["built_in_hash"]),
        // chapter_tree
        .executable(name: "binary_tree", targets: ["binary_tree"]),
        .executable(name: "binary_tree_bfs", targets: ["binary_tree_bfs"]),
        .executable(name: "binary_tree_dfs", targets: ["binary_tree_dfs"]),
        .executable(name: "array_binary_tree", targets: ["array_binary_tree"]),
        .executable(name: "binary_search_tree", targets: ["binary_search_tree"]),
        .executable(name: "avl_tree", targets: ["avl_tree"]),
        // chapter_heap
        .executable(name: "heap", targets: ["heap"]),
        .executable(name: "my_heap", targets: ["my_heap"]),
        .executable(name: "top_k", targets: ["top_k"]),
        // chapter_graph
        .executable(name: "graph_adjacency_matrix", targets: ["graph_adjacency_matrix"]),
        .executable(name: "graph_adjacency_list", targets: ["graph_adjacency_list"]),
        .executable(name: "graph_bfs", targets: ["graph_bfs"]),
        .executable(name: "graph_dfs", targets: ["graph_dfs"]),
        // chapter_searching
        .executable(name: "binary_search", targets: ["binary_search"]),
        .executable(name: "binary_search_insertion", targets: ["binary_search_insertion"]),
        .executable(name: "binary_search_edge", targets: ["binary_search_edge"]),
        .executable(name: "two_sum", targets: ["two_sum"]),
        .executable(name: "linear_search", targets: ["linear_search"]),
        .executable(name: "hashing_search", targets: ["hashing_search"]),
        // chapter_sorting
        .executable(name: "selection_sort", targets: ["selection_sort"]),
        .executable(name: "bubble_sort", targets: ["bubble_sort"]),
        .executable(name: "insertion_sort", targets: ["insertion_sort"]),
        .executable(name: "quick_sort", targets: ["quick_sort"]),
        .executable(name: "merge_sort", targets: ["merge_sort"]),
        .executable(name: "heap_sort", targets: ["heap_sort"]),
        .executable(name: "bucket_sort", targets: ["bucket_sort"]),
        .executable(name: "counting_sort", targets: ["counting_sort"]),
        .executable(name: "radix_sort", targets: ["radix_sort"]),
        // chapter_divide_and_conquer
        .executable(name: "binary_search_recur", targets: ["binary_search_recur"]),
        .executable(name: "build_tree", targets: ["build_tree"]),
        .executable(name: "hanota", targets: ["hanota"]),
        // chapter_backtracking
        .executable(name: "preorder_traversal_i_compact", targets: ["preorder_traversal_i_compact"]),
        .executable(name: "preorder_traversal_ii_compact", targets: ["preorder_traversal_ii_compact"]),
        .executable(name: "preorder_traversal_iii_compact", targets: ["preorder_traversal_iii_compact"]),
        .executable(name: "preorder_traversal_iii_template", targets: ["preorder_traversal_iii_template"]),
        .executable(name: "permutations_i", targets: ["permutations_i"]),
        .executable(name: "permutations_ii", targets: ["permutations_ii"]),
        .executable(name: "subset_sum_i_naive", targets: ["subset_sum_i_naive"]),
        .executable(name: "subset_sum_i", targets: ["subset_sum_i"]),
        .executable(name: "subset_sum_ii", targets: ["subset_sum_ii"]),
        .executable(name: "n_queens", targets: ["n_queens"]),
        // chapter_dynamic_programming
        .executable(name: "climbing_stairs_backtrack", targets: ["climbing_stairs_backtrack"]),
        .executable(name: "climbing_stairs_dfs", targets: ["climbing_stairs_dfs"]),
        .executable(name: "climbing_stairs_dfs_mem", targets: ["climbing_stairs_dfs_mem"]),
        .executable(name: "climbing_stairs_dp", targets: ["climbing_stairs_dp"]),
        .executable(name: "min_cost_climbing_stairs_dp", targets: ["min_cost_climbing_stairs_dp"]),
        .executable(name: "climbing_stairs_constraint_dp", targets: ["climbing_stairs_constraint_dp"]),
        .executable(name: "min_path_sum", targets: ["min_path_sum"]),
        .executable(name: "knapsack", targets: ["knapsack"]),
        .executable(name: "unbounded_knapsack", targets: ["unbounded_knapsack"]),
        .executable(name: "coin_change", targets: ["coin_change"]),
        .executable(name: "coin_change_ii", targets: ["coin_change_ii"]),
        .executable(name: "edit_distance", targets: ["edit_distance"]),
        // chapter_greedy
        .executable(name: "coin_change_greedy", targets: ["coin_change_greedy"]),
        .executable(name: "fractional_knapsack", targets: ["fractional_knapsack"]),
        .executable(name: "max_capacity", targets: ["max_capacity"]),
        .executable(name: "max_product_cutting", targets: ["max_product_cutting"]),
    ],
    dependencies: [
        .package(url: "https://github.com/apple/swift-collections", branch: "release/1.1"),
    ],
    targets: [
        // helper
        .target(name: "utils", path: "utils"),
        .target(name: "graph_adjacency_list_target", dependencies: ["utils"], path: "chapter_graph", sources: ["graph_adjacency_list_target.swift"], swiftSettings: [.define("TARGET")]),
        .target(name: "binary_search_insertion_target", path: "chapter_searching", sources: ["binary_search_insertion_target.swift"], swiftSettings: [.define("TARGET")]),
        // chapter_computational_complexity
        .executableTarget(name: "iteration", path: "chapter_computational_complexity", sources: ["iteration.swift"]),
        .executableTarget(name: "recursion", path: "chapter_computational_complexity", sources: ["recursion.swift"]),
        .executableTarget(name: "time_complexity", path: "chapter_computational_complexity", sources: ["time_complexity.swift"]),
        .executableTarget(name: "worst_best_time_complexity", path: "chapter_computational_complexity", sources: ["worst_best_time_complexity.swift"]),
        .executableTarget(name: "space_complexity", dependencies: ["utils"], path: "chapter_computational_complexity", sources: ["space_complexity.swift"]),
        // chapter_array_and_linkedlist
        .executableTarget(name: "array", path: "chapter_array_and_linkedlist", sources: ["array.swift"]),
        .executableTarget(name: "linked_list", dependencies: ["utils"], path: "chapter_array_and_linkedlist", sources: ["linked_list.swift"]),
        .executableTarget(name: "list", path: "chapter_array_and_linkedlist", sources: ["list.swift"]),
        .executableTarget(name: "my_list", path: "chapter_array_and_linkedlist", sources: ["my_list.swift"]),
        // chapter_stack_and_queue
        .executableTarget(name: "stack", path: "chapter_stack_and_queue", sources: ["stack.swift"]),
        .executableTarget(name: "linkedlist_stack", dependencies: ["utils"], path: "chapter_stack_and_queue", sources: ["linkedlist_stack.swift"]),
        .executableTarget(name: "array_stack", path: "chapter_stack_and_queue", sources: ["array_stack.swift"]),
        .executableTarget(name: "queue", path: "chapter_stack_and_queue", sources: ["queue.swift"]),
        .executableTarget(name: "linkedlist_queue", dependencies: ["utils"], path: "chapter_stack_and_queue", sources: ["linkedlist_queue.swift"]),
        .executableTarget(name: "array_queue", path: "chapter_stack_and_queue", sources: ["array_queue.swift"]),
        .executableTarget(name: "deque", path: "chapter_stack_and_queue", sources: ["deque.swift"]),
        .executableTarget(name: "linkedlist_deque", path: "chapter_stack_and_queue", sources: ["linkedlist_deque.swift"]),
        .executableTarget(name: "array_deque", path: "chapter_stack_and_queue", sources: ["array_deque.swift"]),
        // chapter_hashing
        .executableTarget(name: "hash_map", dependencies: ["utils"], path: "chapter_hashing", sources: ["hash_map.swift"]),
        .executableTarget(name: "array_hash_map", dependencies: ["utils"], path: "chapter_hashing", sources: ["array_hash_map.swift"]),
        .executableTarget(name: "hash_map_chaining", dependencies: ["utils"], path: "chapter_hashing", sources: ["hash_map_chaining.swift"]),
        .executableTarget(name: "hash_map_open_addressing", dependencies: ["utils"], path: "chapter_hashing", sources: ["hash_map_open_addressing.swift"]),
        .executableTarget(name: "simple_hash", path: "chapter_hashing", sources: ["simple_hash.swift"]),
        .executableTarget(name: "built_in_hash", dependencies: ["utils"], path: "chapter_hashing", sources: ["built_in_hash.swift"]),
        // chapter_tree
        .executableTarget(name: "binary_tree", dependencies: ["utils"], path: "chapter_tree", sources: ["binary_tree.swift"]),
        .executableTarget(name: "binary_tree_bfs", dependencies: ["utils"], path: "chapter_tree", sources: ["binary_tree_bfs.swift"]),
        .executableTarget(name: "binary_tree_dfs", dependencies: ["utils"], path: "chapter_tree", sources: ["binary_tree_dfs.swift"]),
        .executableTarget(name: "array_binary_tree", dependencies: ["utils"], path: "chapter_tree", sources: ["array_binary_tree.swift"]),
        .executableTarget(name: "binary_search_tree", dependencies: ["utils"], path: "chapter_tree", sources: ["binary_search_tree.swift"]),
        .executableTarget(name: "avl_tree", dependencies: ["utils"], path: "chapter_tree", sources: ["avl_tree.swift"]),
        // chapter_heap
        .executableTarget(name: "heap", dependencies: ["utils", .product(name: "HeapModule", package: "swift-collections")], path: "chapter_heap", sources: ["heap.swift"]),
        .executableTarget(name: "my_heap", dependencies: ["utils"], path: "chapter_heap", sources: ["my_heap.swift"]),
        .executableTarget(name: "top_k", dependencies: ["utils", .product(name: "HeapModule", package: "swift-collections")], path: "chapter_heap", sources: ["top_k.swift"]),
        // chapter_graph
        .executableTarget(name: "graph_adjacency_matrix", dependencies: ["utils"], path: "chapter_graph", sources: ["graph_adjacency_matrix.swift"]),
        .executableTarget(name: "graph_adjacency_list", dependencies: ["utils"], path: "chapter_graph", sources: ["graph_adjacency_list.swift"]),
        .executableTarget(name: "graph_bfs", dependencies: ["utils", "graph_adjacency_list_target"], path: "chapter_graph", sources: ["graph_bfs.swift"]),
        .executableTarget(name: "graph_dfs", dependencies: ["utils", "graph_adjacency_list_target"], path: "chapter_graph", sources: ["graph_dfs.swift"]),
        // chapter_searching
        .executableTarget(name: "binary_search", path: "chapter_searching", sources: ["binary_search.swift"]),
        .executableTarget(name: "binary_search_insertion", path: "chapter_searching", sources: ["binary_search_insertion.swift"]),
        .executableTarget(name: "binary_search_edge", dependencies: ["binary_search_insertion_target"], path: "chapter_searching", sources: ["binary_search_edge.swift"]),
        .executableTarget(name: "two_sum", path: "chapter_searching", sources: ["two_sum.swift"]),
        .executableTarget(name: "linear_search", dependencies: ["utils"], path: "chapter_searching", sources: ["linear_search.swift"]),
        .executableTarget(name: "hashing_search", dependencies: ["utils"], path: "chapter_searching", sources: ["hashing_search.swift"]),
        // chapter_sorting
        .executableTarget(name: "selection_sort", path: "chapter_sorting", sources: ["selection_sort.swift"]),
        .executableTarget(name: "bubble_sort", path: "chapter_sorting", sources: ["bubble_sort.swift"]),
        .executableTarget(name: "insertion_sort", path: "chapter_sorting", sources: ["insertion_sort.swift"]),
        .executableTarget(name: "quick_sort", path: "chapter_sorting", sources: ["quick_sort.swift"]),
        .executableTarget(name: "merge_sort", path: "chapter_sorting", sources: ["merge_sort.swift"]),
        .executableTarget(name: "heap_sort", path: "chapter_sorting", sources: ["heap_sort.swift"]),
        .executableTarget(name: "bucket_sort", path: "chapter_sorting", sources: ["bucket_sort.swift"]),
        .executableTarget(name: "counting_sort", path: "chapter_sorting", sources: ["counting_sort.swift"]),
        .executableTarget(name: "radix_sort", path: "chapter_sorting", sources: ["radix_sort.swift"]),
        // chapter_divide_and_conquer
        .executableTarget(name: "binary_search_recur", path: "chapter_divide_and_conquer", sources: ["binary_search_recur.swift"]),
        .executableTarget(name: "build_tree", dependencies: ["utils"], path: "chapter_divide_and_conquer", sources: ["build_tree.swift"]),
        .executableTarget(name: "hanota", path: "chapter_divide_and_conquer", sources: ["hanota.swift"]),
        // chapter_backtracking
        .executableTarget(name: "preorder_traversal_i_compact", dependencies: ["utils"], path: "chapter_backtracking", sources: ["preorder_traversal_i_compact.swift"]),
        .executableTarget(name: "preorder_traversal_ii_compact", dependencies: ["utils"], path: "chapter_backtracking", sources: ["preorder_traversal_ii_compact.swift"]),
        .executableTarget(name: "preorder_traversal_iii_compact", dependencies: ["utils"], path: "chapter_backtracking", sources: ["preorder_traversal_iii_compact.swift"]),
        .executableTarget(name: "preorder_traversal_iii_template", dependencies: ["utils"], path: "chapter_backtracking", sources: ["preorder_traversal_iii_template.swift"]),
        .executableTarget(name: "permutations_i", path: "chapter_backtracking", sources: ["permutations_i.swift"]),
        .executableTarget(name: "permutations_ii", path: "chapter_backtracking", sources: ["permutations_ii.swift"]),
        .executableTarget(name: "subset_sum_i_naive", path: "chapter_backtracking", sources: ["subset_sum_i_naive.swift"]),
        .executableTarget(name: "subset_sum_i", path: "chapter_backtracking", sources: ["subset_sum_i.swift"]),
        .executableTarget(name: "subset_sum_ii", path: "chapter_backtracking", sources: ["subset_sum_ii.swift"]),
        .executableTarget(name: "n_queens", path: "chapter_backtracking", sources: ["n_queens.swift"]),
        // chapter_dynamic_programming
        .executableTarget(name: "climbing_stairs_backtrack", path: "chapter_dynamic_programming", sources: ["climbing_stairs_backtrack.swift"]),
        .executableTarget(name: "climbing_stairs_dfs", path: "chapter_dynamic_programming", sources: ["climbing_stairs_dfs.swift"]),
        .executableTarget(name: "climbing_stairs_dfs_mem", path: "chapter_dynamic_programming", sources: ["climbing_stairs_dfs_mem.swift"]),
        .executableTarget(name: "climbing_stairs_dp", path: "chapter_dynamic_programming", sources: ["climbing_stairs_dp.swift"]),
        .executableTarget(name: "min_cost_climbing_stairs_dp", path: "chapter_dynamic_programming", sources: ["min_cost_climbing_stairs_dp.swift"]),
        .executableTarget(name: "climbing_stairs_constraint_dp", path: "chapter_dynamic_programming", sources: ["climbing_stairs_constraint_dp.swift"]),
        .executableTarget(name: "min_path_sum", path: "chapter_dynamic_programming", sources: ["min_path_sum.swift"]),
        .executableTarget(name: "knapsack", path: "chapter_dynamic_programming", sources: ["knapsack.swift"]),
        .executableTarget(name: "unbounded_knapsack", path: "chapter_dynamic_programming", sources: ["unbounded_knapsack.swift"]),
        .executableTarget(name: "coin_change", path: "chapter_dynamic_programming", sources: ["coin_change.swift"]),
        .executableTarget(name: "coin_change_ii", path: "chapter_dynamic_programming", sources: ["coin_change_ii.swift"]),
        .executableTarget(name: "edit_distance", path: "chapter_dynamic_programming", sources: ["edit_distance.swift"]),
        // chapter_greedy
        .executableTarget(name: "coin_change_greedy", path: "chapter_greedy", sources: ["coin_change_greedy.swift"]),
        .executableTarget(name: "fractional_knapsack", path: "chapter_greedy", sources: ["fractional_knapsack.swift"]),
        .executableTarget(name: "max_capacity", path: "chapter_greedy", sources: ["max_capacity.swift"]),
        .executableTarget(name: "max_product_cutting", path: "chapter_greedy", sources: ["max_product_cutting.swift"]),
    ]
)
